from dataclasses import dataclass, field

from typing import Optional, Union, Dict

try:
    from numpy.typing import ArrayLike
except ImportError:
    from numpy import ndarray as ArrayLike

from .float_types import FloatTypes
from .int_types import IntTypes
from .pixel_types import PixelTypes


@dataclass
class PointSetType:
    dimension: int = 3

    pointComponentType: Union[IntTypes, FloatTypes] = FloatTypes.Float32
    pointPixelComponentType: Union[IntTypes, FloatTypes] = FloatTypes.Float32
    pointPixelType: PixelTypes = PixelTypes.Scalar
    pointPixelComponents: int = 1


@dataclass
class PointSet:
    pointSetType: Union[PointSetType, Dict] = field(default_factory=PointSetType)

    name: str = "PointSet"

    numberOfPoints: int = 0
    points: Optional[ArrayLike] = None

    numberOfPointPixels: int = 0
    pointData: Optional[ArrayLike] = None

    metadata: Dict = field(default_factory=dict)

    def __post_init__(self):
        if isinstance(self.pointSetType, dict):
            self.pointSetType = PointSetType(**self.pointSetType)



from typing import Union
import sys

if sys.version_info < (3, 10):
    from importlib_metadata import distribution
else:
    from importlib.metadata import distribution

import numpy as np
from numpy.typing import ArrayLike

from .int_types import IntTypes
from .float_types import FloatTypes
from .to_cupy_array import is_cupy_array

try:
    distribution("dask")
    _DASK_AVAILABLE = True
except:
    _DASK_AVAILABLE = False


def array_like_to_numpy_array(arr: ArrayLike) -> np.ndarray:
    """Convert a numpy array-like to a numpy ndarray.

    :param arr: numpy ndarray like
    :type  arr: ArrayLike

    :return: Numpy array
    :rtype:  np.ndarray
    """
    if isinstance(arr, np.ndarray):
        return arr
    if _DASK_AVAILABLE:
        import dask.array as da

        if isinstance(arr, da.Array):
            arr = arr.compute()
    if is_cupy_array(arr):
        return arr.get()
    return np.array(arr)


def array_like_to_bytes(arr: ArrayLike) -> bytes:
    """Convert a numpy array-like to bytes."""
    if hasattr(arr, "tobytes"):
        return arr.tobytes()
    return array_like_to_numpy_array(arr).tobytes()


def buffer_to_numpy_array(component_type: Union[IntTypes, FloatTypes], buf):
    if component_type == IntTypes.UInt8:
        return np.frombuffer(buf, dtype=np.uint8)
    elif component_type == IntTypes.Int8:
        return np.frombuffer(buf, dtype=np.int8)
    elif component_type == IntTypes.UInt16:
        return np.frombuffer(buf, dtype=np.uint16)
    elif component_type == IntTypes.Int16:
        return np.frombuffer(buf, dtype=np.int16)
    elif component_type == IntTypes.UInt32:
        return np.frombuffer(buf, dtype=np.uint32)
    elif component_type == IntTypes.Int32:
        return np.frombuffer(buf, dtype=np.int32)
    elif component_type == IntTypes.UInt64:
        return np.frombuffer(buf, dtype=np.uint64)
    elif component_type == IntTypes.Int64:
        return np.frombuffer(buf, dtype=np.int64)
    elif component_type == FloatTypes.Float32:
        return np.frombuffer(buf, dtype=np.float32)
    elif component_type == FloatTypes.Float64:
        return np.frombuffer(buf, dtype=np.float64)
    else:
        raise ValueError("Unsupported component type")


def _dtype_to_component_type(dtype):
    if dtype == np.uint8:
        return IntTypes.UInt8
    elif dtype == np.int8:
        return IntTypes.Int8
    elif dtype == np.uint16:
        return IntTypes.UInt16
    elif dtype == np.int16:
        return IntTypes.Int16
    elif dtype == np.uint32:
        return IntTypes.UInt32
    elif dtype == np.int32:
        return IntTypes.Int32
    elif dtype == np.uint64:
        return IntTypes.UInt64
    elif dtype == np.int64:
        return IntTypes.Int64
    elif dtype == np.float32:
        return FloatTypes.Float32
    elif dtype == np.float64:
        return FloatTypes.Float64
    else:
        raise ValueError("Unsupported dtype")



import itk

from itkwasm import PointSet
from dataclasses import asdict
import numpy as np


def test_pointset():
    n_points = 5
    dimension = 3

    PointSetType = itk.PointSet[itk.F, dimension]
    pointset = PointSetType.New()

    points = np.random.random((n_points, dimension)).astype(np.float32)
    pointset.SetPoints(itk.vector_container_from_array(points.ravel()))

    point_data = np.random.random((n_points,)).astype(np.float32)
    pointset.SetPointData(itk.vector_container_from_array(point_data.ravel()))

    itk_pointset_dict = itk.dict_from_pointset(pointset)
    # Bug, to be fixed by 5.3.0
    itk_pointset_dict.pop("dimension", None)
    itkwasm_pointset = PointSet(**itk_pointset_dict)
    itkwasm_pointset_dict = asdict(itkwasm_pointset)
    itk_pointset_roundtrip = itk.pointset_from_dict(itkwasm_pointset_dict)
    itk_pointset_roundtrip_dict = itk.dict_from_pointset(itk_pointset_roundtrip)

    pointSetType = itk_pointset_dict["pointSetType"]
    pointSetType_roundtrip = itk_pointset_roundtrip_dict["pointSetType"]
    assert pointSetType["dimension"] == pointSetType_roundtrip["dimension"]

    assert pointSetType["pointComponentType"] == pointSetType_roundtrip["pointComponentType"]
    assert pointSetType["pointPixelComponentType"] == pointSetType_roundtrip["pointPixelComponentType"]
    assert pointSetType["pointPixelType"] == pointSetType_roundtrip["pointPixelType"]
    assert pointSetType["pointPixelComponents"] == pointSetType_roundtrip["pointPixelComponents"]

    assert itk_pointset_dict["name"] == itk_pointset_roundtrip_dict["name"]

    assert itk_pointset_dict["numberOfPoints"] == itk_pointset_roundtrip_dict["numberOfPoints"]
    assert np.array_equal(itk_pointset_dict["points"], itk_pointset_roundtrip_dict["points"])

    assert itk_pointset_dict["numberOfPointPixels"] == itk_pointset_roundtrip_dict["numberOfPointPixels"]
    assert np.array_equal(itk_pointset_dict["pointData"], itk_pointset_roundtrip_dict["pointData"])



import pytest


def test_function_factory():
    test_accelerator = pytest.importorskip("test_accelerator")

    from itkwasm import function_factory

    interface_package = "itkwasm_example_package"
    func_name = "example_function"
    registered = function_factory.lookup(interface_package, func_name)

    from test_accelerator import slower, faster, fastest

    highest = function_factory.highest_priority(interface_package, func_name)
    assert highest == fastest

    function_factory.set_priority(faster, 30)
    highest = function_factory.highest_priority(interface_package, func_name)
    assert highest == faster

    assert function_factory.get_priority(slower) == -1

    function_factory.disable(interface_package, func_name)
    highest = function_factory.highest_priority(interface_package, func_name)
    assert highest == None


def test_environment_dispatch():
    test_accelerator = pytest.importorskip("test_accelerator")

    from itkwasm import function_factory, environment_dispatch

    interface_package = "itkwasm_example_package"
    func_name = "example_function"

    from test_accelerator import fastest

    function_factory.register(interface_package, func_name, fastest, 1)

    func = environment_dispatch(interface_package, func_name)
    from test_accelerator import fastest

    assert func == fastest



from enum import Enum

class InterfaceTypes(str, Enum):
    TextFile = "TextFile"
    BinaryFile = "BinaryFile"
    TextStream = "TextStream"
    BinaryStream = "BinaryStream"
    Image = "Image"
    Mesh = "Mesh"
    PointSet = "PointSet"
    PolyData = "PolyData"
    Transform = "Transform"
    JsonCompatible = "JsonCompatible"



from dataclasses import dataclass
from typing import Optional, Union

from .interface_types import InterfaceTypes
from .text_file import TextFile
from .text_stream import TextStream
from .binary_file import BinaryFile
from .binary_stream import BinaryStream
from .image import Image
from .mesh import Mesh
from .polydata import PolyData


@dataclass
class PipelineOutput:
    type: InterfaceTypes
    data: Optional[
        Union[
            str,
            bytes,
            TextStream,
            BinaryStream,
            TextFile,
            BinaryFile,
            Image,
            Mesh,
            PolyData,
        ]
    ] = None
    path: Optional[str] = None



from enum import Enum


class PixelTypes(str, Enum):
    Unknown = "Unknown"
    Scalar = "Scalar"
    RGB = "RGB"
    RGBA = "RGBA"
    Offset = "Offset"
    Vector = "Vector"
    Point = "Point"
    CovariantVector = "CovariantVector"
    SymmetricSecondRankTensor = "SymmetricSecondRankTensor"
    DiffusionTensor3D = "DiffusionTensor3D"
    Complex = "Complex"
    FixedArray = "FixedArray"
    Array = "Array"
    Matrix = "Matrix"
    VariableLengthVector = "VariableLengthVector"
    VariableSizeMatrix = "VariableSizeMatrix"

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return f'"{str(self.value)}"'



from dataclasses import dataclass, field

from typing import Optional, Union, Dict

try:
    from numpy.typing import ArrayLike
except ImportError:
    from numpy import ndarray as ArrayLike

from .float_types import FloatTypes
from .int_types import IntTypes
from .pixel_types import PixelTypes


@dataclass
class MeshType:
    dimension: int = 3

    pointComponentType: Union[IntTypes, FloatTypes] = FloatTypes.Float32
    pointPixelComponentType: Union[IntTypes, FloatTypes] = FloatTypes.Float32
    pointPixelType: PixelTypes = PixelTypes.Scalar
    pointPixelComponents: int = 1

    cellComponentType: Union[IntTypes, FloatTypes] = IntTypes.Int32
    cellPixelComponentType: Union[IntTypes, FloatTypes] = FloatTypes.Float32
    cellPixelType: PixelTypes = PixelTypes.Scalar
    cellPixelComponents: int = 1


@dataclass
class Mesh:
    meshType: Union[MeshType, Dict] = field(default_factory=MeshType)

    name: str = "Mesh"

    numberOfPoints: int = 0
    points: Optional[ArrayLike] = None

    numberOfPointPixels: int = 0
    pointData: Optional[ArrayLike] = None

    numberOfCells: int = 0
    cells: Optional[ArrayLike] = None
    cellBufferSize: int = 0

    numberOfCellPixels: int = 0
    cellData: Optional[ArrayLike] = None

    metadata: Dict = field(default_factory=dict)

    def __post_init__(self):
        if isinstance(self.meshType, dict):
            self.meshType = MeshType(**self.meshType)



from enum import Enum


class FloatTypes(str, Enum):
    Float32 = "float32"
    Float64 = "float64"
    SpacePrecisionType = "float64"

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return f'"{str(self.value)}"'



import json
from pathlib import Path, PurePosixPath
from dataclasses import asdict
from typing import List, Union, Dict, Tuple, Set
import ctypes
import sys
import os

import numpy as np

try:
    from numpy.typing import ArrayLike
except ImportError:
    from numpy import ndarray as ArrayLike

from .interface_types import InterfaceTypes
from .pipeline_input import PipelineInput
from .pipeline_output import PipelineOutput
from .text_stream import TextStream
from .binary_stream import BinaryStream
from .text_file import TextFile
from .binary_file import BinaryFile
from .image import Image
from .mesh import Mesh
from .point_set import PointSet
from .polydata import PolyData
from .json_compatible import JsonCompatible
from .int_types import IntTypes
from .float_types import FloatTypes
from .to_numpy_array import (
    buffer_to_numpy_array,
    array_like_to_numpy_array,
    array_like_to_bytes,
)

if sys.platform != "emscripten":
    import zlib
    from platformdirs import user_cache_dir

    from wasmtime import (
        Config,
        Store,
        Engine,
        Module,
        WasiConfig,
        Linker,
        WasmtimeError,
    )

    # Get the value of the ITKWASM_CACHE_DIR environment variable
    # If it is not set, use the default cache directory
    _module_store_dir = os.environ.get("ITKWASM_CACHE_DIR", user_cache_dir("itkwasm"))
    _module_store_dir = Path(_module_store_dir)
    _module_store_dir.mkdir(parents=True, exist_ok=True)


def array_like_to_bytes(arr: ArrayLike) -> bytes:
    """Convert a numpy array-like to bytes."""
    if hasattr(arr, "tobytes"):
        return arr.tobytes()
    return array_like_to_numpy_array(arr).tobytes()


class RunInstance:
    """Helper for working with the wasm module instance created when a Pipeline is run."""

    def __init__(
        self,
        engine: "Engine",
        linker: "Linker",
        module: "Module",
        args: List[str],
        preopen_directories: Set[str],
    ):
        self._store = Store(engine)
        store = self._store

        wasi_config = WasiConfig()
        wasi_config.inherit_env()
        wasi_config.inherit_stderr()
        wasi_config.inherit_stdin()
        wasi_config.inherit_stdout()
        wasi_config.argv = args

        for preopen in preopen_directories:
            wasi_config.preopen_dir(preopen, preopen)

        store.set_wasi(wasi_config)

        instance = linker.instantiate(store, module)
        self._instance = instance

        self._memory = instance.exports(store)["memory"]
        self._input_array_alloc = instance.exports(store)["itk_wasm_input_array_alloc"]
        self._input_json_alloc = instance.exports(store)["itk_wasm_input_json_alloc"]
        self._output_array_address = instance.exports(store)["itk_wasm_output_array_address"]
        self._output_array_size = instance.exports(store)["itk_wasm_output_array_size"]
        self._output_json_address = instance.exports(store)["itk_wasm_output_json_address"]
        self._output_json_size = instance.exports(store)["itk_wasm_output_json_size"]

        _initialize = instance.exports(store)["_initialize"]
        _initialize(store)

    def wasmtime_lift(self, ptr: int, size: int):
        ptr = ptr & 0xFFFFFFFF
        size = size & 0xFFFFFFFF
        if ptr + size > self._memory.data_len(self._store):
            raise IndexError("attempting to lift of bounds")
        raw_base = self._memory.data_ptr(self._store)
        base = ctypes.POINTER(ctypes.c_ubyte)(ctypes.c_ubyte.from_address(ctypes.addressof(raw_base.contents) + ptr))
        return ctypes.string_at(base, size)

    def wasmtime_lower(self, ptr: int, data: Union[bytes, bytearray]):
        ptr = ptr & 0xFFFFFFFF
        size = len(data)
        if ptr + size > self._memory.data_len(self._store):
            raise IndexError("attempting to lower out of bounds")
        raw_base = self._memory.data_ptr(self._store)
        base = ctypes.POINTER(ctypes.c_ubyte)(ctypes.c_ubyte.from_address(ctypes.addressof(raw_base.contents) + ptr))
        ctypes.memmove(base, data, size)

    def set_input_array(self, data_array: Union[bytes, bytearray], input_index: int, sub_index: int) -> int:
        data_ptr = 0
        if data_array != None:
            data_ptr = self._input_array_alloc(self._store, 0, input_index, sub_index, len(data_array))
            self.wasmtime_lower(data_ptr, data_array)
        return data_ptr

    def set_input_json(self, data_object: Dict, input_index: int) -> None:
        data_json = json.dumps(data_object).encode()
        json_ptr = self._input_json_alloc(self._store, 0, input_index, len(data_json))
        self.wasmtime_lower(json_ptr, data_json)

    def get_output_json(self, output_index: int) -> Dict:
        json_ptr = self._output_json_address(self._store, 0, output_index)
        json_len = self._output_json_size(self._store, 0, output_index)
        json_str = self.wasmtime_lift(json_ptr, json_len).decode()
        json_result = json.loads(json_str)
        return json_result

    def get_output_array_address(self, memory: int, output_index: int, output_sub_index: int) -> Dict:
        return self._output_array_address(self._store, memory, output_index, output_sub_index)

    def get_output_array_size(self, memory: int, output_index: int, output_sub_index: int) -> Dict:
        return self._output_array_size(self._store, memory, output_index, output_sub_index)

    def delayed_start(self):
        func = self._instance.exports(self._store)["itk_wasm_delayed_start"]
        return_code = func(self._store)
        return return_code

    def delayed_exit(self, return_code):
        func = self._instance.exports(self._store)["itk_wasm_delayed_exit"]
        func(self._store, return_code)


class Pipeline:
    """Run an itk-wasm WASI pipeline."""

    def __init__(self, pipeline: Union[str, Path, bytes]):
        """Compile the pipeline."""
        self.config = Config()
        self.config.wasm_bulk_memory = True
        self.config.wasm_simd = True
        self.config.wasm_memory64 = True
        self.engine = Engine(self.config)
        if isinstance(pipeline, bytes):
            wasm_bytes = pipeline
        else:
            with open(pipeline, "rb") as fp:
                wasm_bytes = fp.read()
        self.linker = Linker(self.engine)
        self.linker.define_wasi()
        self.linker.allow_shadowing = True
        checksum = zlib.adler32(wasm_bytes)
        module_cache = _module_store_dir / Path(f"module-{checksum}")
        if module_cache.exists():
            try:
                self.module = Module.deserialize_file(self.engine, str(module_cache))
            except WasmtimeError:
                self.module = Module(self.engine, wasm_bytes)
                with module_cache.open("wb") as fp:
                    fp.write(self.module.serialize())
        else:
            self.module = Module(self.engine, wasm_bytes)
            with module_cache.open("wb") as fp:
                fp.write(self.module.serialize())

    def run(
        self,
        args: List[str],
        outputs: List[PipelineOutput] = [],
        inputs: List[PipelineInput] = [],
    ) -> Tuple[PipelineOutput]:
        """Run the itk-wasm pipeline."""

        preopen_directories = set()
        for index, input_ in enumerate(inputs):
            if input_.type == InterfaceTypes.TextFile or input_.type == InterfaceTypes.BinaryFile:
                preopen_directories.add(str(PurePosixPath(input_.data.path).parent))
        for index, output in enumerate(outputs):
            if output.type == InterfaceTypes.TextFile or output.type == InterfaceTypes.BinaryFile:
                preopen_directories.add(str(PurePosixPath(output.data.path).parent))
        preopen_directories = list(preopen_directories)

        ri = RunInstance(self.engine, self.linker, self.module, args, preopen_directories)

        for index, input_ in enumerate(inputs):
            if input_.type == InterfaceTypes.TextStream:
                data_array = input_.data.data.encode()
                array_ptr = ri.set_input_array(data_array, index, 0)
                data_json = {
                    "size": len(data_array),
                    "data": f"data:application/vnd.itk.address,0:{array_ptr}",
                }
                ri.set_input_json(data_json, index)
            elif input_.type == InterfaceTypes.BinaryStream:
                data_array = input_.data.data
                array_ptr = ri.set_input_array(data_array, index, 0)
                data_json = {
                    "size": len(data_array),
                    "data": f"data:application/vnd.itk.address,0:{array_ptr}",
                }
                ri.set_input_json(data_json, index)
            elif input_.type == InterfaceTypes.TextFile:
                pass
            elif input_.type == InterfaceTypes.BinaryFile:
                pass
            elif input_.type == InterfaceTypes.Image:
                image = input_.data
                mv = array_like_to_bytes(image.data)
                data_ptr = ri.set_input_array(mv, index, 0)
                dv = array_like_to_bytes(image.direction)
                direction_ptr = ri.set_input_array(dv, index, 1)
                image_json = {
                    "imageType": asdict(image.imageType),
                    "name": image.name,
                    "origin": image.origin,
                    "spacing": image.spacing,
                    "direction": f"data:application/vnd.itk.address,0:{direction_ptr}",
                    "size": image.size,
                    "data": f"data:application/vnd.itk.address,0:{data_ptr}",
                }
                ri.set_input_json(image_json, index)
            elif input_.type == InterfaceTypes.Mesh:
                mesh = input_.data
                if mesh.numberOfPoints:
                    pv = array_like_to_bytes(mesh.points)
                else:
                    pv = bytes([])
                points_ptr = ri.set_input_array(pv, index, 0)
                if mesh.numberOfCells:
                    cv = array_like_to_bytes(mesh.cells)
                else:
                    cv = bytes([])
                cells_ptr = ri.set_input_array(cv, index, 1)
                if mesh.numberOfPointPixels:
                    pdv = array_like_to_bytes(mesh.pointData)
                else:
                    pdv = bytes([])
                point_data_ptr = ri.set_input_array(pdv, index, 2)
                if mesh.numberOfCellPixels:
                    cdv = array_like_to_bytes(mesh.cellData)
                else:
                    cdv = bytes([])
                cell_data_ptr = ri.set_input_array(cdv, index, 3)
                mesh_json = {
                    "meshType": asdict(mesh.meshType),
                    "name": mesh.name,
                    "numberOfPoints": mesh.numberOfPoints,
                    "points": f"data:application/vnd.itk.address,0:{points_ptr}",
                    "numberOfCells": mesh.numberOfCells,
                    "cells": f"data:application/vnd.itk.address,0:{cells_ptr}",
                    "cellBufferSize": mesh.cellBufferSize,
                    "numberOfPointPixels": mesh.numberOfPointPixels,
                    "pointData": f"data:application/vnd.itk.address,0:{point_data_ptr}",
                    "numberOfCellPixels": mesh.numberOfCellPixels,
                    "cellData": f"data:application/vnd.itk.address,0:{cell_data_ptr}",
                }
                ri.set_input_json(mesh_json, index)
            elif input_.type == InterfaceTypes.PointSet:
                point_set = input_.data
                if point_set.numberOfPoints:
                    pv = array_like_to_bytes(point_set.points)
                else:
                    pv = bytes([])
                points_ptr = ri.set_input_array(pv, index, 0)
                if point_set.numberOfPointPixels:
                    pdv = array_like_to_bytes(point_set.pointData)
                else:
                    pdv = bytes([])
                point_data_ptr = ri.set_input_array(pdv, index, 1)
                point_set_json = {
                    "pointSetType": asdict(point_set.pointSetType),
                    "name": point_set.name,
                    "numberOfPoints": point_set.numberOfPoints,
                    "points": f"data:application/vnd.itk.address,0:{points_ptr}",
                    "numberOfPointPixels": point_set.numberOfPointPixels,
                    "pointData": f"data:application/vnd.itk.address,0:{point_data_ptr}",
                }
                ri.set_input_json(point_set_json, index)
            elif input_.type == InterfaceTypes.PolyData:
                polydata = input_.data
                if polydata.numberOfPoints:
                    pv = array_like_to_bytes(polydata.points)
                else:
                    pv = bytes([])
                points_ptr = ri.set_input_array(pv, index, 0)

                if polydata.verticesBufferSize:
                    pv = array_like_to_bytes(polydata.vertices)
                else:
                    pv = bytes([])
                vertices_ptr = ri.set_input_array(pv, index, 1)

                if polydata.linesBufferSize:
                    pv = array_like_to_bytes(polydata.lines)
                else:
                    pv = bytes([])
                lines_ptr = ri.set_input_array(pv, index, 2)

                if polydata.polygonsBufferSize:
                    pv = array_like_to_bytes(polydata.polygons)
                else:
                    pv = bytes([])
                polygons_ptr = ri.set_input_array(pv, index, 3)

                if polydata.triangleStripsBufferSize:
                    pv = array_like_to_bytes(polydata.triangleStrips)
                else:
                    pv = bytes([])
                triangleStrips_ptr = ri.set_input_array(pv, index, 4)

                if polydata.numberOfPointPixels:
                    pv = array_like_to_bytes(polydata.pointData)
                else:
                    pv = bytes([])
                pointData_ptr = ri.set_input_array(pv, index, 5)

                if polydata.numberOfCellPixels:
                    pv = array_like_to_bytes(polydata.cellData)
                else:
                    pv = bytes([])
                cellData_ptr = ri.set_input_array(pv, index, 6)

                polydata_json = {
                    "polyDataType": asdict(polydata.polyDataType),
                    "name": polydata.name,
                    "numberOfPoints": polydata.numberOfPoints,
                    "points": f"data:application/vnd.itk.address,0:{points_ptr}",
                    "verticesBufferSize": polydata.verticesBufferSize,
                    "vertices": f"data:application/vnd.itk.address,0:{vertices_ptr}",
                    "linesBufferSize": polydata.linesBufferSize,
                    "lines": f"data:application/vnd.itk.address,0:{lines_ptr}",
                    "polygonsBufferSize": polydata.polygonsBufferSize,
                    "polygons": f"data:application/vnd.itk.address,0:{polygons_ptr}",
                    "triangleStripsBufferSize": polydata.triangleStripsBufferSize,
                    "triangleStrips": f"data:application/vnd.itk.address,0:{triangleStrips_ptr}",
                    "numberOfPointPixels": polydata.numberOfPointPixels,
                    "pointData": f"data:application/vnd.itk.address,0:{pointData_ptr}",
                    "numberOfCellPixels": polydata.numberOfCellPixels,
                    "cellData": f"data:application/vnd.itk.address,0:{cellData_ptr}",
                }
                ri.set_input_json(polydata_json, index)
            elif input_.type == InterfaceTypes.JsonCompatible:
                data_array = json.dumps(input_.data).encode()
                array_ptr = ri.set_input_array(data_array, index, 0)
                data_json = {
                    "size": len(data_array),
                    "data": f"data:application/vnd.itk.address,0:{array_ptr}",
                }
                ri.set_input_json(data_json, index)
            else:
                raise ValueError(f"Unexpected/not yet supported input.type {input_.type}")

        return_code = ri.delayed_start()

        populated_outputs: List[PipelineOutput] = []
        if len(outputs) and return_code == 0:
            for index, output in enumerate(outputs):
                output_data = None
                if output.type == InterfaceTypes.TextStream:
                    data_ptr = ri.get_output_array_address(0, index, 0)
                    data_size = ri.get_output_array_size(0, index, 0)
                    data_array = ri.wasmtime_lift(data_ptr, data_size)
                    output_data = PipelineOutput(InterfaceTypes.TextStream, TextStream(data_array.decode()))
                elif output.type == InterfaceTypes.BinaryStream:
                    data_ptr = ri.get_output_array_address(0, index, 0)
                    data_size = ri.get_output_array_size(0, index, 0)
                    data_array = ri.wasmtime_lift(data_ptr, data_size)
                    output_data = PipelineOutput(InterfaceTypes.BinaryStream, BinaryStream(data_array))
                elif output.type == InterfaceTypes.TextFile:
                    output_data = PipelineOutput(InterfaceTypes.TextFile, TextFile(output.data.path))
                elif output.type == InterfaceTypes.BinaryFile:
                    output_data = PipelineOutput(InterfaceTypes.BinaryFile, BinaryFile(output.data.path))
                elif output.type == InterfaceTypes.Image:
                    image_json = ri.get_output_json(index)

                    image = Image(**image_json)

                    data_ptr = ri.get_output_array_address(0, index, 0)
                    data_size = ri.get_output_array_size(0, index, 0)
                    data_array = buffer_to_numpy_array(
                        image.imageType.componentType,
                        ri.wasmtime_lift(data_ptr, data_size),
                    )
                    shape = list(image.size)[::-1]
                    if image.imageType.components > 1:
                        shape.append(image.imageType.components)
                    image.data = data_array.reshape(tuple(shape))

                    direction_ptr = ri.get_output_array_address(0, index, 1)
                    direction_size = ri.get_output_array_size(0, index, 1)
                    direction_array = buffer_to_numpy_array(
                        FloatTypes.Float64,
                        ri.wasmtime_lift(direction_ptr, direction_size),
                    )
                    dimension = image.imageType.dimension
                    direction_array.shape = (dimension, dimension)
                    image.direction = direction_array

                    output_data = PipelineOutput(InterfaceTypes.Image, image)
                elif output.type == InterfaceTypes.Mesh:
                    mesh_json = ri.get_output_json(index)
                    mesh = Mesh(**mesh_json)

                    if mesh.numberOfPoints > 0:
                        data_ptr = ri.get_output_array_address(0, index, 0)
                        data_size = ri.get_output_array_size(0, index, 0)
                        mesh.points = buffer_to_numpy_array(
                            mesh.meshType.pointComponentType,
                            ri.wasmtime_lift(data_ptr, data_size),
                        )
                    else:
                        mesh.points = buffer_to_numpy_array(mesh.meshType.pointComponentType, bytes([]))

                    if mesh.numberOfCells > 0:
                        data_ptr = ri.get_output_array_address(0, index, 1)
                        data_size = ri.get_output_array_size(0, index, 1)
                        mesh.cells = buffer_to_numpy_array(
                            mesh.meshType.cellComponentType,
                            ri.wasmtime_lift(data_ptr, data_size),
                        )
                    else:
                        mesh.cells = buffer_to_numpy_array(mesh.meshType.cellComponentType, bytes([]))
                    if mesh.numberOfPointPixels > 0:
                        data_ptr = ri.get_output_array_address(0, index, 2)
                        data_size = ri.get_output_array_size(0, index, 2)
                        mesh.pointData = buffer_to_numpy_array(
                            mesh.meshType.pointPixelComponentType,
                            ri.wasmtime_lift(data_ptr, data_size),
                        )
                    else:
                        mesh.pointData = buffer_to_numpy_array(mesh.meshType.pointPixelComponentType, bytes([]))

                    if mesh.numberOfCellPixels > 0:
                        data_ptr = ri.get_output_array_address(0, index, 3)
                        data_size = ri.get_output_array_size(0, index, 3)
                        mesh.cellData = buffer_to_numpy_array(
                            mesh.meshType.cellPixelComponentType,
                            ri.wasmtime_lift(data_ptr, data_size),
                        )
                    else:
                        mesh.cellData = buffer_to_numpy_array(mesh.meshType.cellPixelComponentType, bytes([]))

                    output_data = PipelineOutput(InterfaceTypes.Mesh, mesh)
                elif output.type == InterfaceTypes.PointSet:
                    point_set_json = ri.get_output_json(index)
                    point_set = PointSet(**point_set_json)

                    if point_set.numberOfPoints > 0:
                        data_ptr = ri.get_output_array_address(0, index, 0)
                        data_size = ri.get_output_array_size(0, index, 0)
                        point_set.points = buffer_to_numpy_array(
                            point_set.pointSetType.pointComponentType,
                            ri.wasmtime_lift(data_ptr, data_size),
                        )
                    else:
                        point_set.points = buffer_to_numpy_array(point_set.pointSetType.pointComponentType, bytes([]))

                    if point_set.numberOfPointPixels > 0:
                        data_ptr = ri.get_output_array_address(0, index, 1)
                        data_size = ri.get_output_array_size(0, index, 1)
                        point_set.pointData = buffer_to_numpy_array(
                            point_set.pointSetType.pointPixelComponentType,
                            ri.wasmtime_lift(data_ptr, data_size),
                        )
                    else:
                        point_set.pointData = buffer_to_numpy_array(point_set.pointSetType.pointPixelComponentType, bytes([]))

                    output_data = PipelineOutput(InterfaceTypes.PointSet, point_set)
                elif output.type == InterfaceTypes.PolyData:
                    polydata_json = ri.get_output_json(index)
                    polydata = PolyData(**polydata_json)

                    if polydata.numberOfPoints > 0:
                        data_ptr = ri.get_output_array_address(0, index, 0)
                        data_size = ri.get_output_array_size(0, index, 0)
                        polydata.points = buffer_to_numpy_array(
                            FloatTypes.Float32, ri.wasmtime_lift(data_ptr, data_size)
                        )
                    else:
                        polydata.points = buffer_to_numpy_array(FloatTypes.Float32, bytes([]))

                    if polydata.verticesBufferSize > 0:
                        data_ptr = ri.get_output_array_address(0, index, 1)
                        data_size = ri.get_output_array_size(0, index, 1)
                        polydata.vertices = buffer_to_numpy_array(
                            IntTypes.UInt32, ri.wasmtime_lift(data_ptr, data_size)
                        )
                    else:
                        polydata.vertices = buffer_to_numpy_array(IntTypes.UInt32, bytes([]))

                    if polydata.linesBufferSize > 0:
                        data_ptr = ri.get_output_array_address(0, index, 2)
                        data_size = ri.get_output_array_size(0, index, 2)
                        polydata.lines = buffer_to_numpy_array(IntTypes.UInt32, ri.wasmtime_lift(data_ptr, data_size))
                    else:
                        polydata.lines = buffer_to_numpy_array(IntTypes.UInt32, bytes([]))

                    if polydata.polygonsBufferSize > 0:
                        data_ptr = ri.get_output_array_address(0, index, 3)
                        data_size = ri.get_output_array_size(0, index, 3)
                        polydata.polygons = buffer_to_numpy_array(
                            IntTypes.UInt32, ri.wasmtime_lift(data_ptr, data_size)
                        )
                    else:
                        polydata.polygons = buffer_to_numpy_array(IntTypes.UInt32, bytes([]))

                    if polydata.triangleStripsBufferSize > 0:
                        data_ptr = ri.get_output_array_address(0, index, 4)
                        data_size = ri.get_output_array_size(0, index, 4)
                        polydata.triangleStrips = buffer_to_numpy_array(
                            IntTypes.UInt32, ri.wasmtime_lift(data_ptr, data_size)
                        )
                    else:
                        polydata.triangleStrips = buffer_to_numpy_array(IntTypes.UInt32, bytes([]))

                    if polydata.numberOfPointPixels > 0:
                        data_ptr = ri.get_output_array_address(0, index, 5)
                        data_size = ri.get_output_array_size(0, index, 5)
                        polydata.pointData = buffer_to_numpy_array(
                            polydata.polyDataType.pointPixelComponentType,
                            ri.wasmtime_lift(data_ptr, data_size),
                        )
                    else:
                        polydata.triangleStrips = buffer_to_numpy_array(
                            polydata.polyDataType.pointPixelComponentType, bytes([])
                        )

                    if polydata.numberOfCellPixels > 0:
                        data_ptr = ri.get_output_array_address(0, index, 6)
                        data_size = ri.get_output_array_size(0, index, 6)
                        polydata.cellData = buffer_to_numpy_array(
                            polydata.polyDataType.cellPixelComponentType,
                            ri.wasmtime_lift(data_ptr, data_size),
                        )
                    else:
                        polydata.triangleStrips = buffer_to_numpy_array(
                            polydata.polyDataType.cellPixelComponentType, bytes([])
                        )

                    output_data = PipelineOutput(InterfaceTypes.PolyData, polydata)
                elif output.type == InterfaceTypes.JsonCompatible:
                    data_ptr = ri.get_output_array_address(0, index, 0)
                    data_size = ri.get_output_array_size(0, index, 0)
                    data_array = ri.wasmtime_lift(data_ptr, data_size)
                    output_data = PipelineOutput(InterfaceTypes.JsonCompatible, json.loads(data_array.decode()))
                else:
                    raise ValueError(f"Unexpected/not yet supported output.type {output.type}")

                populated_outputs.append(output_data)

        ri.delayed_exit(return_code)

        # Should we be returning the return_code?
        return tuple(populated_outputs)



import pytest
import sys

if sys.version_info < (3, 10):
    pytest.skip("Skipping pyodide tests on older Python", allow_module_level=True)

from pytest_pyodide import run_in_pyodide

from itkwasm import __version__ as test_package_version


@pytest.fixture
def package_wheel():
    return f"itkwasm-{test_package_version}-py3-none-any.whl"


@run_in_pyodide(packages=["micropip", "numpy"])
async def test_image_conversion(selenium, package_wheel):
    import micropip

    await micropip.install(package_wheel)

    from itkwasm import Image
    from itkwasm.pyodide import to_js, to_py
    import numpy as np

    image = Image()

    assert image.imageType.dimension == 2
    assert image.imageType.componentType == "uint8"
    assert image.imageType.pixelType == "Scalar"
    assert image.imageType.components == 1

    assert image.name == "Image"
    assert image.origin[0] == 0.0
    assert image.origin[1] == 0.0
    assert image.spacing[0] == 1.0
    assert image.spacing[1] == 1.0
    assert np.array_equal(image.direction, np.eye(2).astype(np.float64))

    assert image.size[0] == 1
    assert image.size[1] == 1
    image.size = [4, 4]

    assert isinstance(image.metadata, dict)
    image.metadata["a_string"] = "some text"
    image.metadata["an int"] = 3
    assert image.data == None
    image.data = np.arange(16, dtype=np.uint8).reshape((4, 4))
    image_js = to_js(image)

    image_py = to_py(image_js)
    assert image_py.imageType.dimension == 2
    assert image_py.imageType.componentType == "uint8"
    assert image_py.imageType.pixelType == "Scalar"
    assert image_py.imageType.components == 1

    assert image_py.name == "Image"
    assert image_py.origin[0] == 0.0
    assert image_py.origin[1] == 0.0
    assert image_py.spacing[0] == 1.0
    assert image_py.spacing[1] == 1.0
    assert np.array_equal(image_py.direction, np.eye(2).astype(np.float64))

    assert image_py.size[0] == 4
    assert image_py.size[1] == 4

    assert image_py.metadata["a_string"] == "some text"
    assert image_py.metadata["an int"] == 3

    assert isinstance(image_py.metadata, dict)
    assert np.array_equal(image_py.data, np.arange(16, dtype=np.uint8).reshape((4, 4)))


@run_in_pyodide(packages=["micropip", "numpy"])
async def test_point_set_conversion(selenium, package_wheel):
    import micropip

    await micropip.install(package_wheel)

    from itkwasm import PointSet, PointSetType, PixelTypes, FloatTypes
    from itkwasm.pyodide import to_js, to_py
    import numpy as np

    n_points = 5
    dimension = 3

    points = np.random.random((n_points, dimension)).astype(np.float32)
    point_data = np.random.random((n_points,)).astype(np.float32)

    point_set_type = PointSetType(
        dimension,
        FloatTypes.Float32,
        FloatTypes.Float32,
        PixelTypes.Scalar,
        FloatTypes.Float32,
    )

    point_set = PointSet(point_set_type, "point_set", n_points, points, n_points, point_data)

    point_set_js = to_js(point_set)
    point_set_py = to_py(point_set_js)

    point_set_type_py = point_set_py.pointSetType
    assert point_set_type.dimension == point_set_type_py.dimension
    assert point_set_type.pointComponentType == point_set_type_py.pointComponentType
    assert point_set_type.pointPixelComponentType == point_set_type_py.pointPixelComponentType
    assert point_set_type.pointPixelType == point_set_type_py.pointPixelType
    assert point_set_type.pointPixelComponents == point_set_type_py.pointPixelComponents

    assert point_set.name == point_set_py.name
    assert point_set.numberOfPoints == point_set_py.numberOfPoints
    assert np.array_equal(point_set.points, point_set_py.points)
    assert point_set.numberOfPointPixels == point_set_py.numberOfPointPixels
    assert np.array_equal(point_set.pointData, point_set_py.pointData)


@run_in_pyodide(packages=["micropip", "numpy"])
async def test_mesh_conversion(selenium, package_wheel):
    import micropip

    await micropip.install(package_wheel)

    from itkwasm import Mesh, MeshType
    from itkwasm.pyodide import to_js, to_py
    import numpy as np

    n_points = 5
    dimension = 3

    mesh_type = MeshType()

    points = np.random.random((n_points, dimension)).astype(np.float32)
    point_data = np.random.random((n_points,)).astype(np.float32)

    mesh = Mesh(
        mesh_type,
        points=points,
        numberOfPoints=n_points,
        pointData=point_data,
        numberOfPointPixels=n_points,
    )

    mesh_js = to_js(mesh)
    mesh_py = to_py(mesh_js)

    mesh_type_py = mesh_py.meshType
    assert mesh_type.dimension == mesh_type_py.dimension
    assert mesh_type.pointComponentType == mesh_type_py.pointComponentType
    assert mesh_type.pointPixelComponentType == mesh_type_py.pointPixelComponentType
    assert mesh_type.pointPixelType == mesh_type_py.pointPixelType
    assert mesh_type.pointPixelComponents == mesh_type_py.pointPixelComponents

    assert mesh.name == mesh_py.name
    assert mesh.numberOfPoints == mesh_py.numberOfPoints
    assert np.array_equal(mesh.points, mesh_py.points)
    assert mesh.numberOfPointPixels == mesh_py.numberOfPointPixels
    assert np.array_equal(mesh.pointData, mesh_py.pointData)


@run_in_pyodide(packages=["micropip", "numpy"])
async def test_polydata_conversion(selenium, package_wheel):
    import micropip

    await micropip.install(package_wheel)

    from itkwasm import PolyData, PolyDataType
    from itkwasm.pyodide import to_js, to_py
    import numpy as np

    n_points = 5
    dimension = 3

    polydata_type = PolyDataType()

    points = np.random.random((n_points, dimension)).astype(np.float32)
    point_data = np.random.random((n_points,)).astype(np.float32)

    polydata = PolyData(
        polydata_type,
        points=points,
        numberOfPoints=n_points,
        pointData=point_data,
        numberOfPointPixels=n_points,
    )

    polydata_js = to_js(polydata)
    polydata_py = to_py(polydata_js)

    polydata_type_py = polydata_py.polyDataType
    assert polydata_type.pointPixelComponentType == polydata_type_py.pointPixelComponentType
    assert polydata_type.pointPixelType == polydata_type_py.pointPixelType
    assert polydata_type.pointPixelComponents == polydata_type_py.pointPixelComponents

    assert polydata.name == polydata_py.name
    assert polydata.numberOfPoints == polydata_py.numberOfPoints
    assert np.array_equal(polydata.points, polydata_py.points)
    assert polydata.numberOfPointPixels == polydata_py.numberOfPointPixels
    assert np.array_equal(polydata.pointData, polydata_py.pointData)


@run_in_pyodide(packages=["micropip", "numpy"])
async def test_binary_stream_conversion(selenium, package_wheel):
    import micropip

    await micropip.install(package_wheel)

    from itkwasm import BinaryStream
    from itkwasm.pyodide import to_js, to_py

    data = bytes([222, 173, 190, 239])
    binary_stream = BinaryStream(data)

    binary_stream_js = to_js(binary_stream)
    binary_stream_py = to_py(binary_stream_js)

    assert binary_stream_py.data[0], 222
    assert binary_stream_py.data[1], 173
    assert binary_stream_py.data[2], 190
    assert binary_stream_py.data[3], 239


@run_in_pyodide(packages=["micropip", "numpy"])
async def test_text_stream_conversion(selenium, package_wheel):
    import micropip

    await micropip.install(package_wheel)

    from itkwasm import TextStream
    from itkwasm.pyodide import to_js, to_py

    data = "The answer is 42."
    text_stream = TextStream(data)

    text_stream_js = to_js(text_stream)
    text_stream_py = to_py(text_stream_js)

    assert text_stream_py.data == data


@run_in_pyodide(packages=["micropip", "numpy"])
async def test_binary_file_conversion(selenium, package_wheel):
    import micropip

    await micropip.install(package_wheel)

    from itkwasm import BinaryFile
    from itkwasm.pyodide import to_js, to_py
    import numpy as np
    from pathlib import PurePosixPath

    data = bytes([222, 173, 190, 239])
    path = PurePosixPath("file.bin")
    with open(path, "wb") as fp:
        fp.write(data)
    binary_file = BinaryFile(path)

    binary_file_js = to_js(binary_file)
    binary_file_py = to_py(binary_file_js)

    with open(binary_file_py.path, "rb") as fp:
        data_py = fp.read()

    assert data_py[0], 222
    assert data_py[1], 173
    assert data_py[2], 190
    assert data_py[3], 239


@run_in_pyodide(packages=["micropip", "numpy"])
async def test_text_file_conversion(selenium, package_wheel):
    import micropip

    await micropip.install(package_wheel)

    from itkwasm import TextFile
    from itkwasm.pyodide import to_js, to_py
    import numpy as np
    from pathlib import PurePosixPath

    data = "The answer is 42."
    path = PurePosixPath("file.txt")
    with open(path, "w") as fp:
        fp.write(data)
    text_file = TextFile(path)

    text_file_js = to_js(text_file)
    text_file_py = to_py(text_file_js)

    with open(text_file_py.path, "r") as fp:
        data_py = fp.read()

    assert data_py == data


@run_in_pyodide(packages=["micropip", "numpy"])
async def test_list_conversion(selenium, package_wheel):
    import micropip

    await micropip.install(package_wheel)

    from itkwasm import TextFile
    from itkwasm.pyodide import to_js, to_py
    import numpy as np
    from pathlib import PurePosixPath

    data = "The answer is 42."

    def create_text_file(index):
        path = PurePosixPath(f"file{index}.txt")
        with open(path, "w") as fp:
            fp.write(data)
        text_file = TextFile(path)
        return text_file

    text_files = [create_text_file(index) for index in range(4)]

    text_files_js = to_js(text_files)
    text_files_py = to_py(text_files_js)

    def verify_text_file(text_file):
        with open(text_file.path, "r") as fp:
            data_py = fp.read()

        assert data_py == data

    for text_file in text_files_py:
        verify_text_file(text_file)


@run_in_pyodide(packages=["micropip", "numpy"])
async def test_uint8array_conversion(selenium, package_wheel):
    import micropip

    await micropip.install(package_wheel)

    from itkwasm.pyodide import to_js, to_py

    data = bytes([222, 173, 190, 239])

    data_js = to_js(data)
    data_py = to_py(data_js)

    assert isinstance(data_py, bytes)

    assert data_py[0], 222
    assert data_py[1], 173
    assert data_py[2], 190
    assert data_py[3], 239

    ivalue = 8
    ivalue_js = to_js(ivalue)
    ivalue_py = to_py(ivalue_js)

    assert ivalue == ivalue_py


@run_in_pyodide(packages=["micropip"])
async def test_json_object_conversion(selenium, package_wheel):
    import micropip

    await micropip.install(package_wheel)

    from itkwasm.pyodide import to_js, to_py

    data = {"a": 1, "b": True, "c": {"nested": True}}

    data_js = to_js(data)
    data_py = to_py(data_js)

    assert isinstance(data_py, dict)

    assert data_py["a"] == 1
    assert data_py["b"] == True
    assert data_py["c"]["nested"] == True



"""itkwasm: Python interface to itk-wasm WebAssembly modules."""

__version__ = "1.0b180"

from .interface_types import InterfaceTypes
from .image import Image, ImageType, ImageRegion
from .point_set import PointSet, PointSetType
from .mesh import Mesh, MeshType
from .transform import TransformList, Transform, TransformType, TransformParameterizations
from .polydata import PolyData, PolyDataType
from .binary_file import BinaryFile
from .binary_stream import BinaryStream
from .text_file import TextFile
from .text_stream import TextStream
from .json_compatible import JsonCompatible
from .pipeline import Pipeline
from .pipeline_input import PipelineInput
from .pipeline_output import PipelineOutput
from .float_types import FloatTypes
from .int_types import IntTypes
from .pixel_types import PixelTypes
from .environment_dispatch import environment_dispatch, function_factory
from .cast_image import cast_image
from .image_from_array import image_from_array
from .to_numpy_array import (
    array_like_to_numpy_array,
    array_like_to_bytes,
    buffer_to_numpy_array,
)
from .to_cupy_array import array_like_to_cupy_array

__all__ = [
    "InterfaceTypes",
    "Pipeline",
    "PipelineInput",
    "PipelineOutput",
    "Image",
    "ImageType",
    "ImageRegion",
    "PointSet",
    "PointSetType",
    "Mesh",
    "MeshType",
    "PolyData",
    "PolyDataType",
    "TransformList",
    "Transform",
    "TransformType",
    "TransformParameterizations",
    "BinaryFile",
    "BinaryStream",
    "TextFile",
    "TextStream",
    "JsonCompatible",
    "FloatTypes",
    "IntTypes",
    "PixelTypes",
    "environment_dispatch",
    "function_factory",
    "cast_image",
    "image_from_array",
    "array_like_to_numpy_array",
    "array_like_to_bytes",
    "array_like_to_cupy_array",
    "buffer_to_numpy_array",
]



from dataclasses import dataclass
from typing import Optional, Union

from .interface_types import InterfaceTypes
from .text_file import TextFile
from .text_stream import TextStream
from .binary_file import BinaryFile
from .binary_stream import BinaryStream
from .image import Image
from .mesh import Mesh
from .polydata import PolyData


@dataclass
class PipelineInput:
    type: InterfaceTypes
    data: Union[
        str,
        bytes,
        TextStream,
        BinaryStream,
        TextFile,
        BinaryFile,
        Image,
        Mesh,
        PolyData,
    ]
    path: Optional[str] = None



from dataclasses import dataclass, field
from enum import Enum
from typing import Sequence, Union, Dict, Optional, List

try:
    from numpy.typing import ArrayLike
except ImportError:
    from numpy import ndarray as ArrayLike
import numpy as np

class TransformParameterizations(str, Enum):
    Composite = "Composite"
    Identity = "Identity"
    Translation = "Translation"
    Euler2D = "Euler2D"
    Euler3D = "Euler3D"
    Rigid2D = "Rigid2D"
    Rigid3D = "Rigid3D"
    Rigid3DPerspective = "Rigid3DPerspective"
    Versor = "Versor"
    VersorRigid3D = "VersorRigid3D"
    Scale = "Scale"
    ScaleLogarithmic = "ScaleLogarithmic"
    ScaleSkewVersor3D = "ScaleSkewVersor3D"
    Similarity2D = "Similarity2D"
    Similarity3D = "Similarity3D"
    QuaternionRigid = "QuaternionRigid"
    Affine = "Affine"
    ScalableAffine = "ScalableAffine"
    AzimuthElevationToCartesian = "AzimuthElevationToCartesian"
    BSpline = "BSpline"
    BSplineSmoothingOnUpdateDisplacementField = "BSplineSmoothingOnUpdateDisplacementField"
    ConstantVelocityField = "ConstantVelocityField"
    DisplacementField = "DisplacementField"
    GaussianSmoothingOnUpdateDisplacementField = "GaussianSmoothingOnUpdateDisplacementField"
    GaussianExponentialDiffeomorphic = "GaussianExponentialDiffeomorphic"
    VelocityField = "VelocityField"
    TimeVaryingVelocityField = "TimeVaryingVelocityField"
    GaussianSmoothingOnUpdateTimeVaryingVelocityField = "GaussianSmoothingOnUpdateTimeVaryingVelocityField"

@dataclass
class TransformType:
    inputDimension: int = 3
    outputDimension: int = 3
    transformParameterization: TransformParameterizations = TransformParameterizations.Identity

@dataclass
class Transform:
    transformType: Union[TransformType, Dict] = field(default_factory=TransformType)
    numberOfFixedParameters: int = 0
    numberOfParameters: int = 0

    name: str = "Transform"

    inputSpaceName: str = ""
    outputSpaceName: str = ""

    fixedParameters: Optional[ArrayLike] = None
    parameters: Optional[ArrayLike] = None

    metadata: Dict = field(default_factory=dict)

    def __post_init__(self):
        if isinstance(self.transformType, dict):
            self.transformType = TransformType(**self.transformType)

TransformList = List[Transform]